unit ICMakeUniqueName;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module provides a function for making unique names as per the Finder.
	ie If you have a name like "foo", the function will return "foo copy",
	"foo copy 1", and so on.
	
	This module is partly WorldScript aware.  Some functions, such as string
	comparison, are done in a WorldScript savvy fashion, but others (such
	as string breaking and string to decimal coversion, are not.  Each
	WorldScript failure point is marked with a comment  WorldScript ,
	for easy identification and, hopefully one day, correction.
*)

interface

	uses
		Types;
		
	const
		kMUCompareAsFilenames = 1;
		kMUCompareAsScriptStrings = 2;
	type
		(*	Forward declaration of the pointer type to allow it as a parameter
			to MakeNameUniqueFunction.
		*)

		MakeNameUniqueParamsPtr = ^MakeNameUniqueParams;
		
		(*	This is a callback function that the client must supply to
			MakeUniqueEqualString.  MakeUniqueEqualString calls this function
			after it has generated a candidate unique name to determine whether
			the name is actually unique.
			
			The function should return noErr if the testName is unique, 1
			if testName is not unique, or some negative error code if the test
			could not be performed for some reason.  If it returns a negative
			error code, MakeUniqueEqualString will also return that error code.
		*)

		MakeNameUniqueFunction = function (params : MakeNameUniqueParamsPtr; const testName : Str255) : OSStatus;

		(* See MakeNameUnique for a description of the following type. *)
		
		MakeNameUniqueParams =
			record
				comparisonType : integer;
				script : ScriptCode;
				lang : ScriptCode;
				lengthLimit : integer;
				copyStr : Str255;
				copyIStr : Str255;
				uniqueFunction : MakeNameUniqueFunction;
				refcon : Ptr;
			end;

	function MakeUniqueEqualString(const params : MakeNameUniqueParams; const str1 : Str255; const str2 : Str255) : Boolean;
		(*	A utility routine which your uniqueFunction callback can call to
			compare strings.  This routine return true if str1 is equal to
			str2 (case insensitive) as determined by the comparision algorithm
			defined in params.
		*)

	function MakeNameUnique(const startName : Str255;
							const params : MakeNameUniqueParams;
							var uniqueName : Str255) : OSStatus;
		(*	Given a starting name, and a bunch of parameters (which are
			typically unchanged between multiple calls to this routine), this
			routine returns a unique name based on startName.  The params must
			be set up as follows:
			
			o comparisonType should be set to either kMUCompareAsFilenames (in
			  in which case the routine acts as if the names are file names
			  and compares them as the File Manager would compare them) or
			  kMUCompareAsScriptStrings (in which case the routine acts as if
			  the names are in the script system defined by script and lang.

			o script and lang define the script system for the startName
			  copyStr, copyIStr, and uniqueName.  These values have no meaning
			  if comparisonType is kMUCompareAsFilenames.
			
			o lengthLimit defines the maximum length of the returned uniqueName.
			  It must be less than 255, and enough longer than copyStr and copyIStr
			  to allow some sort of unique name to be formed.
			
			o copyStr and copyIStr define how uniqueName is formed.  If the
			  startName doesn't contain a suffix of copyStr or copyIStr + X (where
			  X is some integer), then copyStr is appended to the name to form
			  the first candidate uniqueName.  If startName does contain a suffix,
			  copyIStr + X (with an incremented X) is used to form the unique name.
			  In English, you pass copyStr as ' copy' and copyIStr as ' copy '.
			
			o uniqueFunction is called back by this routine to determine whether
			  a candidate name that it has generated is unique.  See the description
			  of the MakeNameUniqueFunction for a description.
			  
			o refcon is provided for the benefit of the callback function.
		*)
	
	function FileManagerUniqueFunction(params : MakeNameUniqueParamsPtr; const testName : Str255) : OSStatus;
		(*	If you're trying to make a unique file name, then you can use this
			function as the uniqueFunction you pass to MakeNameUnique.  If you do
			this, you must set up refcon to point to the FSSpec from which
			startName is derived.  In addition, you must set comparisonType to
			kMUCompareAsFilenames and lengthLimit to 31.  This function will
			determine whether testName is unique in the directory denoted by
			FSSpecPtr(refcon)^.vRefNum + FSSpecPtr(refcon)^.parID.
		*)
	
implementation

	uses
		TextUtils,
		Script,

		ICDebug,		
		ICCommonSubs;

	function AppendStringsWithLengthLimit(const leftStr : Str255; const rightStr : Str255; lengthLimit : integer) : Str255;
		(*	This function appends rightStr to leftStr while keeping the total
			string length less than lengthLimit.  leftStr is truncated to achieve
			this.
			
			WorldScript  The reduction in length of the leftStr should
			only be done on character boundaries, ie if we're going to remove the
			second byte of a two byte character, we should remove the entire
			character.
		*)
		var
			tmpStr : Str255;
	begin
		ICAssert(@leftStr <> nil);
		ICAssert(@rightStr <> nil);
		ICAssert(lengthLimit > length(rightStr));
		ICAssert(lengthLimit <= 255);
		
		if length(leftStr) + length(rightStr) > lengthLimit then begin
			tmpStr := concat(copy(leftStr, 1, lengthLimit - length(rightStr)), rightStr);
		end else begin
			tmpStr := concat(leftStr, rightStr);
		end; (* if *)
		AppendStringsWithLengthLimit := tmpStr;
	end; (* AppendStringsWithLengthLimit *)
	
	procedure PositionAndValueOfNumber(const str : Str255; var position : integer; var number : longint);
		(*	This function scans str backwards looking for a numeric suffix.  If there
			is one, it returns number as the value of the numeric suffix, and position
			as the index of the last non-numeric character in str.  [position can
			also be considered the length of the non-numeric prefix string.]
			It sets number to 0 and position to length(str) if there is no numeric 
			suffix.
			
			WorldScript  We should step by character by character, not byte
			by byte.  Also, we should use Script Manager to tell whether a character
			is a number.  Also, we should use StringToExtended instead of DecVal.
		*)
		var
			numStr : Str255;
	begin
		numStr := '';
		position := length(str);

		// Limit the length of numStr to 9, because 999,999,999
		// is the maximum power of 10 - 1 that we can represent
		// in a longint.
		
		while (position > 0) & (length(numStr) < 9) & (str[position] in ['0'..'9']) do begin
			numStr := concat(str[position], numStr);
			position := position - 1;
		end; (* while *)
		if numStr = '' then begin
			number := 0;
		end else begin
			number := DecVal(numStr);
		end; (* if *)
	end; (* PositionAndValueOfNumber *)
	
	function MakeUniqueEqualString(const params : MakeNameUniqueParams; const str1 : Str255; const str2 : Str255) : Boolean;
		(*	See comment in interface part. *)
		var
			result : Boolean;
			temp1 : Str255;
			temp2 : Str255;
	begin
		case params.comparisonType of
			kMUCompareAsFilenames :
				begin
					result := EqualString(str1, str2, false, true);
				end;
			kMUCompareAsScriptStrings :
				begin
					temp1 := str1;
					temp2 := str2;
					UppercaseText(@temp1[1], length(temp1), params.script);
					UppercaseText(@temp2[1], length(temp2), params.script);
					result := (temp1[0] = temp2[0]) & 
										(TextOrder(@temp1[1], @temp2[1],
												length(temp1), length(temp2),
												params.script, params.script,
												params.lang, params.lang) = 0);
				end;
		end; (* case *)
		MakeUniqueEqualString := result;
	end; (* MakeUniqueEqualString *)
	
	function EqualStringSlice(const params : MakeNameUniqueParams;
								const str1 : Str255; position1 : integer;
								const str2 : Str255; position2 : integer; 
								len : integer) : Boolean;
		(*	This routine extracts a range of characters from each string,
			and compares that range for equality.
		*)
		var
			slice1 : Str255;
			slice2 : Str255;
	begin
		slice1 := copy(str1, position1, len);
		slice2 := copy(str2, position2, len);
		EqualStringSlice :=  MakeUniqueEqualString(params, slice1, slice2);
	end; (* EqualStringSlice *)
	
	procedure CalculateBaseName(const startName : Str255; const params : MakeNameUniqueParams;
								var baseName : Str255; var number : longint);
		(*	This routine finds baseName for a given startName.  The baseName
			is defined as the startName minus copyStr, or baseName minus
			copyIStr + X.  The routine also returns number as the X, or as 0
			if there is no X.
		*)
		var
			position : integer;
	begin
		baseName := startName;
		
		// First start by looking for a number at the end of startName.
		
		PositionAndValueOfNumber(startName, position, number);
		if position = length(startName) then begin
		
			// There is no number, check to see if startName
			// ends with copyStr.

			if length(startName) > length(params.copyStr) then begin
				if EqualStringSlice(params, startName, position - length(params.copyStr) + 1, 
										params.copyStr, 1, 
										length(params.copyStr)) then begin

					// It does.  Return baseName as startName minus copyStr.
					
					baseName := copy(startName, 1, length(startName) - length(params.copyStr));
				end; (* if *)
			end; (* if *)
			
		end else begin

			// There is a number at the, check to see whether it's preceeded
			// by copyIStr.
			
			if position > length(params.copyIStr) then begin
				if EqualStringSlice(params, startName, position - length(params.copyIStr) + 1,
										params.copyIStr, 1,
										length(params.copyIStr)) then begin

					// If is.  Return baseName as startName minus copyIStr + X.
					
					baseName := copy(startName, 1, position - length(params.copyIStr));
				end; (* if *)
			end; (* if *)
			
		end; (* if *)
	end; (* CalculateBaseName *)
	
	function MakeNameUnique(const startName : Str255;
							const params : MakeNameUniqueParams;
							var uniqueName : Str255) : OSStatus;
		(* See comments in interface part.

			WorldScript  Should use ExtendedToString instead of DecStr.
		*)
		var
			err : OSStatus;
			baseName : Str255;
			number : longint;
	begin
		ICAssert(params.lengthLimit >= length(params.copyStr));
		ICAssert(params.lengthLimit >= length(params.copyIStr));
		ICAssert(params.lengthLimit <= 255);
		ICAssert(params.uniqueFunction <> nil);

		err := 1;
		
		// Start by checking the startNmae.
		// If that's unique, there's nothing more to do.  If it isn't,
		// we have some work to do.
		
		uniqueName := startName;
		err := params.uniqueFunction(@params, uniqueName);

		// Now attempt to form a unique name using "startName" + " copy".
			
		if err = 1 then begin
		
			CalculateBaseName(startName, params, baseName, number);
			
			if length(baseName) = length(startName) then begin
				
				// baseName is the same as startName, ie there is no
				// copyStr suffix, or copyIStr + X suffix.  See whether
				// startName + copyStr is unique.
				
				uniqueName := AppendStringsWithLengthLimit(startName, 
														params.copyStr, 
														params.lengthLimit);
				err := params.uniqueFunction(@params, uniqueName);
			end; (* if *)
		end; (* if *)
			
		// If neither of the above worked, start forming names like
		// "startName" + " copy " + X
			
		if err = 1 then begin
			repeat
				number := number + 1;
				uniqueName := AppendStringsWithLengthLimit(baseName, 
														concat(params.copyIStr, DecStr(number)), 
														params.lengthLimit);
				err := params.uniqueFunction(@params, uniqueName);
			until err <> 1;
		end; (* if *)
		
		MakeNameUnique := err;
	end; (* MakeNameUnique *)

	function FileManagerUniqueFunction(params : MakeNameUniqueParamsPtr; const testName : Str255) : OSStatus;
		(*	See comment in interface part. *)
		var
			err : OSStatus;
			testFSSpec : FSSpec;
			cpb : CInfoPBRec;
	begin
		ICAssert(params^.refcon <> nil);
		ICAssert(params^.comparisonType = kMUCompareAsFilenames);
		ICAssert(params^.lengthLimit = 31);
		ICAssert(length(testName) <= 31);
		
		testFSSpec := FSSpecPtr(params^.refcon)^;
		testFSSpec.name := testName;
		
		err := FSpGetCatInfo(testFSSpec, 0, cpb);
		if err = noErr then begin
			err := 1;				// testName is not unique
		end else if (err = fnfErr) then begin
			err := noErr;			// testName is unique
		end else begin
			// All other errors are returned back as is.
		end; (* if *)
		
		FileManagerUniqueFunction := err;
	end; (* FileManagerUniqueFunction *)

end. (* ICMakeUniqueName *)
