unit ICSystemGlobals;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This file contains code for managing the Internet Config system-wide globals.
*)

interface

	uses
		Types,
		Files,
		
		InternetConfig;
		
	const
		gestaltInternetConfigGlobals = 'ICAp';

	(* System-Wide Globals
	   -------------------
	   These globals are registered once per system start, typically
	   by the component's "Register" routine.  They hold important
	   information that is global to all instances of all components.
	*)
	
	const
		kICDebugComponentEntryMask    = $0001;
		kICDebugComponentExitMask     = $0002;
		kICDebugSignificantEventsMask = $0004;
	
	{$ifc icDebug}
		{$definec ComponentWatchPoint(globals, mask, message, param) ComponentWatchPointProc(globals, mask, message, param)}
	{$elsec}
		{$definec ComponentWatchPoint(globals, mask, message, param) }
	{$endc}

{$options align=mac68k}
	type
		ICSystemGlobals =
			record
				magic : OSType;			// should always be gestaltInternetConfigGlobals
				componentVersion : NumVersion;
				globalsSize : longint;
				debugFlags : longint;
				gestaltProcPtr : Ptr;

				primaryConfigFile : FSSpec;
				primarySeed : longint;
				
				// Fields for interrupt safe API routines
				
				mappings : Ptr;
				mappingsSize : longint;
				interruptSafeCount : longint;
			end;
		ICSystemGlobalsPtr = ^ICSystemGlobals;
{$options align=reset}

	(* ***** Core Data Structure ***** *)

	(* The ICRRecord is the core data structure behind IC.  It holds all
		of the information needed to operate an instance, ie a connection
		to the IC system.
		
		The ICRRecord is instantiated in two different places.  Firstly,
		each component instance has a copy of this data structure.  The
		fields in the data structure are all used.  Most importantly,
		the instance field is used to hold the component's "target" instance.

		If the application talks directly to the component API (ie ICCAPI.p)
		then an ICInstance is a ComponentInstance and this is the
		only ICRRecord that comes into play.

		However, if an application uses the 68K glue and doesn't talk
		directly to the component, the 68K glue creates a copy of this
		record too.  This copy is used in one of two ways: if the component
		is installed, ICRRecord.instance contains the component instance
		for it and the other fields of the record are ignored.  If the
		component isn't installed, ICRRecord.instance is nil and the
		glue's link-in implementation comes into play.  The other fields
		in ICRRecord are then used to hold the link-in implementation's
		instance variables.
		
		The other part of the IC equation is the switch glue.  This is the
		glue that the application actually calls (in ICAPI.p) and this
		contains a bunch of routines that go:
		
			if ICRRecord.instance is nil
				call the link-in implementation
			else
				call the component
			end-if
		
		The fact that the component stores its target in its
		ICRRecord.instance field is a subtle but critical implementation
		detail.  Components, when they call themselves (ie a component
		routine that calls another component routine that is exported),
		are supposed to call through the current target.  This allows
		another component to set itself as the target and change or
		capture specific behaviour.
		
		For example, say I want to change the behaviour of ICParseURL.
		I could write an override component that reimplements
		ICParseURL and all client calls to ICParseURL would come to
		me for modification.  But when the client calls ICLaunchURL
		and the component's internal implementation calls ICParseURL,
		it won't necessarily come to me.  In order to make this work, the
		component is required to make these internal calls through the target.
		
		IC does this by having the link-in implementation call the IC API
		(ICAPI.p) when it needs to call itself.  Because a pointer to
		an ICRRecord is a valid ICInstance, it can just do this by
		taking the address of the inst parameter and casting it to
		an ICInstance.  Now there are two cases.  If the link-in
		implementation is being run stand-alone, ie it's hard-linked
		into an application and is not using the component, the
		ICRRecord.instance pointer will be nil and the switch glue will
		route the call back to the link-in implementation.  Alternatively,
		if the link-in implementation is being run as part of the
		component, the ICRRecord.instance will be the component's
		target, and the switch glue will call that target with the
		routines in ICCAPI.p.  Normally this will be the component
		itself, and we'll end up back in the component's link-in
		implementation.  But if the component has been overridden, we'll
		go to the override component.
		
		Cool huh?
		
		The ICRRecord is *completely* private to the implementation!  Do not rely
		on the existence or placement of these fields.
	*)
		
	(*	Component Globals
		-----------------
	
		The component maintains very few globals variables.  The most important
		global is "inst", a instance of the ICRRecord.  See the comment in ICLinkIn.p
		for more information about the ICRRecord.
		
		Note that the component uses the "instance" field of ICRRecord as a
		"target" variable, ie the place it maintains the component's current target.
		
		The other fields are a "self" reference (mostly used to pass back to
		various Component Manager routines that require a reference to
		the component) and a local copy of the componentVersion.
	*)
	
	type
		ProfileCachePtr = ^array [1..1000] of ResType;
		ProfileCacheHandle = ^ProfileCachePtr;

	type
		ComponentGlobals =
			record
				(* Component info, set up by ICCIOpen. *)
				self: ComponentInstance;
				target: ComponentInstance;

				(* The rest are inited by ICCIStart, except where noted. *)
				
				sysGlobals : ICSystemGlobalsPtr;		// System-wide globals
				
				(* Various flags. *)
				haveStarted : Boolean;
				haveConfigFile: Boolean;
				haveDefaultConfig : Boolean;
				waitingForFirstBegin: Boolean;
				thisFileModified: Boolean;
				registeredAsInterruptSafe : Boolean;

				(* Info about the current config file. *)
				configFileSpec: FSSpec;
				configFileRefNum: integer;
				configFilePerm: ICPerm;
				configProfileResType : ResType;		(* inited by Begin *)

				(* Info about the component's resource file. *)
				oldResourceFileRefNum : integer;
				resourceFileRefNum : integer;

				(* Static info taken from the component's resource file. *)
				componentVersion: NumVersion;
				defaultFileName: Str63;
				icApplicationCreator : OSType;
				
				(* Profile cache, built lazily by profile routines, disposed in ICCIEnd. *)
				profileCache : ProfileCacheHandle;
			end;
		ComponentGlobalsPtr = ^ComponentGlobals;

	(* The fields of the ICRRecord include:
	
		target -- See the description above.
		haveConfigFile -- A Boolean that says whether this instance has been
			configured, ie we have decided on a definitive config file.  Specifically
			this means that configFileSpec is valid.
		waitingForFirstBegin -- A Boolean that says whether someone has
			called Begin on this config.  This is set every time we change
			config files, and cleared every time Begin is called.  The component
			uses this Boolean to tell whether this is the first Begin, in which
			cases it sanity checks the config file before calling through
			to the basic implementation.
		thisFileModified -- A Boolean that says whether this config file has
			been modified.  This is set every time SetPref is called, and
			cleared every time a new config file is specified.  The component
			uses this to determine whether it should save a new backup of
			this configuration file.
		configFileSpec -- A file spec that points to the current configuration
			file.  Only valid if haveConfigFile is set.
		configFileRefNum -- A Resource Manager refNum for the current
			config file.  This is 0 if the file isn't open.  When someone
			calls Begin, the file is opened and this is set to the file refNum.
			When they call End, the opposite happens.
		configFilePerm -- The access permissions on the current file.  If it's not
			icNoPerm, you're inside a Begin/End pair.
		defaultFileName -- The default file name for the Internet
			Preferences file.  This is stored in the instance for historical
			reasons.  In the current environment, this code should simply
			call GetDefaultFileName to get it.
	*)

	(* File Format Notes
		----------------
		The Internet Preferences file format is very simple.  IC is basically
		a thin layer over the Resource Manager.  All IC preferences are
		stored as resources of type kICPrefResourceTypeBase.  The key for
		the preference is stored as the resource.  The attributes for
		the preference are stored as the first 4 bytes of the resource.
		The resource ID is not significant.
		
		In IC 1.4 or later, the file also profits a version number, which
		is the version number of the component that created the default
		preferences in the file.  This version number is not set up by
		the link-in glue.  If a future component successfully updates the
		default preferences, this version will be set to the component's
		version.
	*)

	function FindICSystemGlobals(globals: ComponentGlobalsPtr) : ICSystemGlobalsPtr;
		(* Return the system-wide globals, or nil if they haven't been registered
			yet.
		*)

	function RegisterSystemGlobals(globals: ComponentGlobalsPtr) : OSStatus;
		(* Create and register the system wide globals. *)

	{$ifc icDebug}
		procedure ComponentWatchPointProc(globals: ComponentGlobalsPtr; debugFlagMask : longint; const message : Str255; param : longint);
	{$endc}
	
implementation

	uses
		GestaltEqu,
		Memory,
		Resources,
		Folders,
		Traps,
		
		ICDebug,
		ICStringLists,
		ICCommonSubs,
		ICComponentResources;

	function FindICSystemGlobals(globals: ComponentGlobalsPtr) : ICSystemGlobalsPtr;
		var
			result : ICSystemGlobalsPtr;
	begin
		{$unused globals}
		if Gestalt(gestaltInternetConfigGlobals, longint(result)) <> noErr then begin
			result := nil;
		end; (* if *)
		FindICSystemGlobals := result;
	end; (* FindICSystemGlobals *)

	function CreateSystemGlobals(globals: ComponentGlobalsPtr; var sysGlobals : ICSystemGlobalsPtr) : OSStatus;
		(*	Creates the IC system-wide globals in the system heap,
			fills out the easy fields, and returns the pointer
			to the caller in the sysGlobals parameters.  If the
			pointer is set by this routine, the caller is still
			responsible for disposing it, even if the routine
			returns an error.
			
			Assumes that the IC resources are first in the chain.
		*)
		var
			err : OSStatus;
			versH : VersRecHndl;
	begin
		{$unused globals}
		err := noErr;
		sysGlobals := ICSystemGlobalsPtr(NewPtrSysClear(sizeof(ICSystemGlobals)));
		if sysGlobals = nil then begin
			err := MemError();
			ICAssert(err <> noErr);
		end; (* if *)
		if err = noErr then begin
			sysGlobals^.magic := gestaltInternetConfigGlobals;
			sysGlobals^.globalsSize := sizeof(ICSystemGlobals);
			sysGlobals^.primarySeed := 0;
		end; (* if *)
		if err = noErr then begin
			versH := VersRecHndl(Get1Resource('vers', 1));
			err := CheckResError(versH);
			if err = noErr then begin
				sysGlobals^.componentVersion := versH^^.numericVersion;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder, 
									sysGlobals^.primaryConfigFile.vRefNum, 
									sysGlobals^.primaryConfigFile.parID);
			if err = noErr then begin
				sysGlobals^.primaryConfigFile.name := GetIndStr(rMiscStrings, strDefaultFileName);
				ICAssert(sysGlobals^.primaryConfigFile.name <> '');
			end; (* if *)
		end; (* if *)
		CreateSystemGlobals := err;
	end; (* CreateSystemGlobals *)
	
	function LoadGestaltProc(globals: ComponentGlobalsPtr; sysGlobals : ICSystemGlobalsPtr) : OSStatus;
		(*	Loads the component's Gestalt proc from a 'PROC' resource
			in the component's resource file.  I first load the resource,
			then create a pointer to hold a copy of it, then copy the code
			into that pointer.  This is because a) I never trust the Resource
			Manager to load things in the right heap, and b) I want the code
			to be locked low in the system heap.
			
			Assumes that the IC resources are first in the chain.
		*)
		var
			err : OSStatus;
			gestaltProcHandle : Handle;
			gestaltProcSize : longint;
	begin
		{$unused globals}
		gestaltProcHandle := Get1Resource('PROC', rGestaltProc);
		err := CheckResError(gestaltProcHandle);
		if err = noErr then begin
			gestaltProcSize := GetHandleSize(gestaltProcHandle);
			sysGlobals^.gestaltProcPtr := NewPtrSys(gestaltProcSize);
			err := MemError();
		end; (* if *)
		if err = noErr then begin
			BlockMoveData(gestaltProcHandle^, sysGlobals^.gestaltProcPtr, gestaltProcSize);
			LongIntPtr(longint(sysGlobals^.gestaltProcPtr) + kGestaltProcDataOffset)^ := longint(sysGlobals);
			MakeDataExecutableAs68KCode(sysGlobals^.gestaltProcPtr, gestaltProcSize);
		end; (* if *)
		LoadGestaltProc := err;
	end; (* LoadGestaltProc *)

	(*	FragRegisterFileLibs is a CFM SPI (System Programming Interface)
		that is meant to be private to Apple.  If unregister is false, it
		registers the file specified by fss as a searchable file, ie
		CFM will search it when it's looking for libraries.  See the
		comment in RegisterComponentAsSharedLibraryContainer (below)
		for a description of why this is necessary.
		
		I feel this is OK to call even though it's a private routine
		because, working at DTS, I've learnt how many developers are
		using it (either because DTS told them about it, or because they
		worked it out for themselves).
	*)
	
	function FragRegisterFileLibs(var fss : FSSpec; unregister : Boolean) : OSErr;
		{$ifc not GENERATINGCFM}
		inline
			$3f3c, $ffec, 		// move.w #$ffec,-(sp)
			$aa5a;				// _CodeFragmentDispatch
		{$elsec}
		external;
		{$endc}
	
	function RegisterComponentAsSharedLibraryContainer(globals: ComponentGlobalsPtr;
									var haveRegistered : Boolean) : OSStatus;
		(*	This routine calls CFM to register the component's data fork as a CFM
			shared library.  This way, our CFM based clients will be able to find
			the IC libraries in our data fork.  [Normally CFM only searches files
			of type 'shlb', but IC must necessarily be of type 'thng', so we have
			to do this trickery.
			
			Assumes that the IC resources are first in the chain.
		*)
		var
			err : OSStatus;
			extensionFileSpec : FSSpec;
	begin
		{$unused globals}
		err := noErr;
		if TrapAvailable(_CodeFragmentDispatch) then begin
			err := GetFSSpecGivenFileRefNum(CurResFile, extensionFileSpec);
			if err = noErr then begin
				err := FragRegisterFileLibs(extensionFileSpec, false);
				haveRegistered := (err = noErr);
			end; (* if *)
		end; (* if *)
		RegisterComponentAsSharedLibraryContainer := err;
	end; (* RegisterComponentAsSharedLibraryContainer *)

	function RegisterSystemGlobals(globals: ComponentGlobalsPtr) : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			junk : OSStatus;
			sysGlobals : ICSystemGlobalsPtr;
			haveRegistered : Boolean;
			didOpen : Boolean;
	begin
		ICAssert(globals <> nil);
		ICAssert(FindICSystemGlobals(globals) = nil);

		sysGlobals := nil;
		haveRegistered := false;
		
		err := StartUsingComponentResources(globals, didOpen);
		if err = noErr then begin

			err := CreateSystemGlobals(globals, sysGlobals);
			if err = noErr then begin
				err := LoadGestaltProc(globals, sysGlobals);
			end; (* if *)
			if err = noErr then begin
				err := RegisterComponentAsSharedLibraryContainer(globals, haveRegistered);
			end; (* if *)
			if err = noErr then begin
				err := NewGestalt(gestaltInternetConfigGlobals, sysGlobals^.gestaltProcPtr);
				if err = noErr then begin
					// All is successful.  Clear out the pointers so that
					// they're not cleaned up at the end of this routine.
					sysGlobals := nil;
				end; (* if *)
			end; (* if *)

		end; (* if *)
		
		err2 := StopUsingComponentResources(globals, didOpen);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		// Clean up.
		
		if (err <> noErr) and haveRegistered then begin
			ICAssert(sysGlobals <> nil);
			junk := FragRegisterFileLibs(sysGlobals^.primaryConfigFile, true);
			ICAssert(junk = noErr);
		end; (* if *)
		if sysGlobals <> nil then begin
			if sysGlobals^.gestaltProcPtr <> nil then begin
				DisposePtr(sysGlobals^.gestaltProcPtr);
				ICAssert(MemError() = noErr);
			end; (* if *)
			DisposePtr(Ptr(sysGlobals));
			ICAssert(MemError() = noErr);
		end; (* if *)

		RegisterSystemGlobals := err;
	end; (* RegisterSystemGlobals *)

	{$ifc icDebug}

		procedure ComponentWatchPointProc(globals: ComponentGlobalsPtr; debugFlagMask : longint; const message : Str255; param : longint);
			var
				sysGlobals : ICSystemGlobalsPtr;
		begin
			ICAssert(globals <> nil);
			sysGlobals := FindICSystemGlobals(globals);
			ICAssert(sysGlobals <> nil);
			
			if band(sysGlobals^.debugFlags, debugFlagMask) <> 0 then begin
				DebugStr('ComponentWatchPointProc: ' + message + ' ' + DecStr(param));
			end; (* if *)
		end; (* ComponentWatchPointProc *)
		
	{$endc}

end. (* ICSystemGlobals *)